@@ -1,6 +1,7 @@
 #define _GNU_SOURCE
 #include <assert.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <getopt.h>
 #include <poll.h>
 #include <signal.h>
@@ -9,8 +10,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
-#include <dbus-1.0/dbus/dbus.h>
+#include <sys/stat.h>
+
+#include <dbus/dbus.h>
 
 #include "version.h"
 
@@ -61,33 +65,36 @@ usage(const char *progname)
         "       %s --server-capabilities\n\n"
         ""
         "Options:\n"
-        "  -a,--app-name=NAME          app-name, typically used for the icon, unless\n"
-        "                              --icon is used\n"
-        "  -i,--icon=ICON              notification icon. Can be either a symbolic name,\n"
-        "                              or a file (default: none)\n"
-        "  -u,--urgency=URGENCY        notification urgency; low, normal or critical\n"
-        "                              (default: normal)\n"
-        "  -c,--category=CATEGORY      notification category (default: none)\n"
-        "  -A,--action=NAME:LABEL      defines an action to display. Implies --wait. May\n"
-        "                              be set multiple times. The name of the\n"
-        "                              triggered action is output to stdout\n"
-        "  -H,--hint=TYPE:NAME:VALUE   specifies additional hints. Valid types are\n"
-        "                              boolean, byte, int, double and string\n"
-        "  -r,--replaces=ID            update an existing notification\n"
-        "  -t,--expire-time=TIME_MS    notification timeout, in milliseconds (default:\n"
-        "                              server defined)\n"
-        "     --transient              show a transient notification. Transient\n"
-        "                              notifications by-pass the server's persistence\n"
-        "                              capability, if any.\n"
-        "  -C,--close=ID               close an existing notification\n"
-        "  -p,--print-id               print the notification ID on stdout\n"
-        "  -R,--print-reason           print the reason the notification was closed\n"
-        "  -T,--print-token            print the activation token, if present\n"
-        "  -w,--wait                   wait for the notification to be closed before\n"
-        "                              exiting\n"
-        "     --server-info            display server name and version\n"
-        "     --server-capabilities    display server capabilities\n"
-        "  -v,--version                show the version number and quit\n"
+        "  -a,--app-name=NAME           app-name, typically used for the icon, unless\n"
+        "                               --icon is used\n"
+        "  -i,--icon=ICON               notification icon. Can be either a symbolic name,\n"
+        "                               or a file (default: none)\n"
+        "     --image-data=FILE         file content (must be raw RGBA pixel data) is sent\n"
+        "                               as a image-data hint\n"
+        "     --image-size=WIDTHxHEIGHT width and height, in pixels, of --image-data\n"
+        "  -u,--urgency=URGENCY         notification urgency; low, normal or critical\n"
+        "                               (default: normal)\n"
+        "  -c,--category=CATEGORY       notification category (default: none)\n"
+        "  -A,--action=NAME:LABEL       defines an action to display. Implies --wait. May\n"
+        "                               be set multiple times. The name of the\n"
+        "                               triggered action is output to stdout\n"
+        "  -H,--hint=TYPE:NAME:VALUE    specifies additional hints. Valid types are\n"
+        "                               boolean, byte, int, double and string\n"
+        "  -r,--replaces=ID             update an existing notification\n"
+        "  -t,--expire-time=TIME_MS     notification timeout, in milliseconds (default:\n"
+        "                               server defined)\n"
+        "     --transient               show a transient notification. Transient\n"
+        "                               notifications by-pass the server's persistence\n"
+        "                               capability, if any.\n"
+        "  -C,--close=ID                close an existing notification\n"
+        "  -p,--print-id                print the notification ID on stdout\n"
+        "  -R,--print-reason            print the reason the notification was closed\n"
+        "  -T,--print-token             print the activation token, if present\n"
+        "  -w,--wait                    wait for the notification to be closed before\n"
+        "                               exiting\n"
+        "     --server-info             display server name and version\n"
+        "     --server-capabilities     display server capabilities\n"
+        "  -v,--version                 show the version number and quit\n"
         , progname, progname, progname, progname);
 }
 
@@ -109,7 +116,7 @@ send_close_notification(DBusConnection *conn, uint32_t notification_id)
     /* Mimic notify-send: close the notification */
     DBusMessageIter args;
     dbus_message_iter_init_append(msg, &args);
-    dbus_message_iter_append_basic(&args, 'u', &notification_id);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &notification_id);
     dbus_connection_send(conn, msg, 0);
     dbus_connection_flush(conn);
     dbus_message_unref(msg);
@@ -318,6 +325,124 @@ out:
     return ret;
 }
 
+static void
+add_urgency_hint(DBusMessageIter *hints, enum urgency urgency)
+{
+    static const char *name = "urgency";
+
+    DBusMessageIter hint;
+    DBusMessageIter value;
+    dbus_message_iter_open_container(hints, DBUS_TYPE_DICT_ENTRY, 0, &hint);
+
+    dbus_message_iter_append_basic(&hint, DBUS_TYPE_STRING, &name);
+    dbus_message_iter_open_container(&hint, DBUS_TYPE_VARIANT, DBUS_TYPE_BYTE_AS_STRING, &value);
+    dbus_message_iter_append_basic(&value, DBUS_TYPE_BYTE, &(uint8_t){urgency});
+    dbus_message_iter_close_container(&hint, &value);
+    dbus_message_iter_close_container(hints, &hint);
+}
+
+static void
+add_category_hint(DBusMessageIter *hints, const char *category)
+{
+    static const char *name = "category";
+
+    if (category == NULL || category[0] == '\0')
+        return;
+
+    DBusMessageIter hint;
+    DBusMessageIter value;
+    dbus_message_iter_open_container(hints, DBUS_TYPE_DICT_ENTRY, 0, &hint);
+    dbus_message_iter_append_basic(&hint, DBUS_TYPE_STRING, &name);
+    dbus_message_iter_open_container(&hint, DBUS_TYPE_VARIANT, DBUS_TYPE_STRING_AS_STRING, &value);
+    dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &category);
+    dbus_message_iter_close_container(&hint, &value);
+    dbus_message_iter_close_container(hints, &hint);
+}
+
+static void
+add_transient_hint(DBusMessageIter *hints, bool transient)
+{
+    static const char *name = "transient";
+
+    if (!transient)
+        return;
+
+    DBusMessageIter hint;
+    DBusMessageIter value;
+
+    dbus_message_iter_open_container(hints, DBUS_TYPE_DICT_ENTRY, 0, &hint);
+    dbus_message_iter_append_basic(&hint, DBUS_TYPE_STRING, &name);
+    dbus_message_iter_open_container(&hint, DBUS_TYPE_VARIANT, DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+    dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &(dbus_bool_t){transient});
+    dbus_message_iter_close_container(&hint, &value);
+    dbus_message_iter_close_container(hints, &hint);
+}
+
+static void
+add_image_data_hint(DBusMessageIter *hints, const uint8_t *image,
+                    int32_t width, int32_t height)
+{
+    static const char *name = "image-data";
+
+    if (width == 0 || height == 0)
+        return;
+
+    DBusMessageIter hint;
+    DBusMessageIter value;
+
+    dbus_message_iter_open_container(hints, DBUS_TYPE_DICT_ENTRY, 0, &hint);
+    dbus_message_iter_append_basic(&hint, DBUS_TYPE_STRING, &name);
+
+    dbus_message_iter_open_container(
+        &hint, DBUS_TYPE_VARIANT,
+        DBUS_STRUCT_BEGIN_CHAR_AS_STRING
+          DBUS_TYPE_INT32_AS_STRING
+          DBUS_TYPE_INT32_AS_STRING
+          DBUS_TYPE_INT32_AS_STRING
+          DBUS_TYPE_BOOLEAN_AS_STRING
+          DBUS_TYPE_INT32_AS_STRING
+          DBUS_TYPE_INT32_AS_STRING
+          DBUS_TYPE_ARRAY_AS_STRING
+            DBUS_TYPE_BYTE_AS_STRING
+        DBUS_STRUCT_END_CHAR_AS_STRING,
+        &value);
+
+    /* We only support RGBA image data */
+    const int32_t stride = width * 4;
+
+    DBusMessageIter data;
+    dbus_message_iter_open_container(&value, DBUS_TYPE_STRUCT, NULL, &data);
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_INT32, &width);
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_INT32, &height);
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_INT32, &stride);
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_BOOLEAN, &(dbus_bool_t){true});
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_INT32, &(int32_t){8});
+    dbus_message_iter_append_basic(&data, DBUS_TYPE_INT32, &(int32_t){4});
+
+    DBusMessageIter pixels;
+    dbus_message_iter_open_container(&data, DBUS_TYPE_ARRAY,DBUS_TYPE_BYTE_AS_STRING, &pixels);
+    for (size_t i = 0; i < height * stride; i++)
+        dbus_message_iter_append_basic(&pixels, DBUS_TYPE_BYTE, &image[i]);
+    dbus_message_iter_close_container(&data, &pixels);
+    dbus_message_iter_close_container(&value, &data);
+    dbus_message_iter_close_container(&hint, &value);
+    dbus_message_iter_close_container(hints, &hint);
+}
+
+static void
+add_user_hint(DBusMessageIter *hints, const struct hint *h)
+{
+    DBusMessageIter hint;
+    DBusMessageIter value;
+
+    dbus_message_iter_open_container(hints, DBUS_TYPE_DICT_ENTRY, 0, &hint);
+    dbus_message_iter_append_basic(&hint, DBUS_TYPE_STRING, &h->name);
+    dbus_message_iter_open_container(&hint, DBUS_TYPE_VARIANT, h->type_as_string, &value);
+    dbus_message_iter_append_basic(&value, h->type, &h->u);
+    dbus_message_iter_close_container(&hint, &value);
+    dbus_message_iter_close_container(hints, &hint);
+}
+
 int
 main(int argc, char *const *argv)
 {
@@ -329,10 +454,14 @@ main(int argc, char *const *argv)
 #define OPT_SERVER_INFO         256
 #define OPT_SERVER_CAPABILITIES 257
 #define OPT_TRANSIENT           258
+#define OPT_IMAGE_DATA          259
+#define OPT_IMAGE_SIZE          260
 
     const struct option longopts[] = {
         {"app-name",            required_argument, NULL, 'a'},
         {"icon",                required_argument, NULL, 'i'},
+        {"image-data",          required_argument, NULL, OPT_IMAGE_DATA},
+        {"image-size",          required_argument, NULL, OPT_IMAGE_SIZE},
         {"urgency",             required_argument, NULL, 'u'},
         {"category",            required_argument, NULL, 'c'},
         {"action",              required_argument, NULL, 'A'},
@@ -354,8 +483,9 @@ main(int argc, char *const *argv)
     };
 
     const char *progname = argv[0];
-    const char *app_id = program_invocation_name;
+    const char *app_id = progname;
     const char *icon = NULL;
+    const char *image_data_file = NULL;
     const char *category = NULL;
 
     uint32_t replaces_id = 0;
@@ -379,6 +509,9 @@ main(int argc, char *const *argv)
 
     char *body = NULL;
     char *icon_uri = NULL;
+    uint8_t *image_data = NULL;
+    int32_t image_width = 0;
+    int32_t image_height = 0;
 
     DBusError err = DBUS_ERROR_INIT;
     DBusMessage *msg = NULL;
@@ -467,6 +600,10 @@ main(int argc, char *const *argv)
                 hint.type = DBUS_TYPE_STRING;
                 hint.type_as_string = DBUS_TYPE_STRING_AS_STRING;
                 hint.u.string = value;
+
+                /* Ignore empty string values */
+                if (value[0] == '\0')
+                    break;
             } else {
                 fprintf(stderr, "error: invalid hint type: %s\n", type_str);
                 goto out;
@@ -538,6 +675,32 @@ main(int argc, char *const *argv)
             break;
         }
 
+        case OPT_IMAGE_DATA:
+            image_data_file = optarg;
+            break;
+
+        case OPT_IMAGE_SIZE: {
+            errno = 0;
+            char *end = NULL;
+            const unsigned long w = strtoul(optarg, &end, 10);
+
+            if (errno == 0 && end != NULL && *end == 'x' && *(end + 1) != '\0') {
+                const unsigned long h = strtoul(end + 1, &end, 10);
+
+                if (errno == 0 && end != NULL && *end == '\0') {
+                    image_width = w;
+                    image_height = h;
+                } else {
+                    fprintf(stderr, "error: %s: invalid image size\n", optarg);
+                    goto out;
+                }
+            } else {
+                fprintf(stderr, "error: %s: invalid image size\n", optarg);
+                goto out;
+            }
+            break;
+        }
+
         case OPT_SERVER_INFO:
             server_info = true;
             break;
@@ -563,18 +726,17 @@ main(int argc, char *const *argv)
     }
 
     const char *title = argv[0];
-    uint32_t transient_value = transient; /* D-Bus "boolean" */
-    uint8_t urgency_value = urgency;      /* D-Bus "byte" */
 
     for (size_t i = 1, len = 0; i < argc; i++) {
         const char *word = argv[i];
         size_t word_length = strlen(word);
+        bool add_space = i > 1;
 
-        body = realloc(body, len + 1 + word_length + 1);
+        body = realloc(body, len + add_space + word_length + 1);
         body[len] = ' ';
-        memcpy(&body[len + 1], word, word_length);
+        memcpy(&body[len + add_space], word, word_length);
 
-        len += 1 + word_length;
+        len += add_space + word_length;
         body[len] = '\0';
     }
 
@@ -619,6 +781,44 @@ main(int argc, char *const *argv)
         goto out;
     }
 
+    if (image_data_file != NULL) {
+        struct stat st;
+        int fd = open(image_data_file, O_RDONLY);
+
+        if (fd < 0 || fstat(fd, &st) < 0) {
+            fprintf(stderr, "error: %s: failed to open+stat: %s\n",
+                    image_data_file, strerror(errno));
+            if (fd >= 0)
+                close(fd);
+            goto out;
+        }
+
+        /* 4 bytes per pixel (RGBA) */
+        size_t expected_size = image_width * image_height * 4;
+        if (st.st_size != expected_size) {
+            fprintf(
+                stderr,
+                "error: %s: file size (%ld) does not match image dimensions (%dx%d)\n",
+                image_data_file, (long)st.st_size, image_width, image_height);
+            close(fd);
+            goto out;
+        }
+
+        //printf("allocating %lu image data bytes\n", st.st_size);
+        image_data = malloc(st.st_size);
+
+        if (image_data == NULL ||
+            read(fd, image_data, st.st_size) != (ssize_t)st.st_size)
+        {
+            fprintf(stderr, "error: %s: failed to read: %s\n",
+                    image_data_file, strerror(errno));
+            close(fd);
+            goto out;
+        }
+
+        close(fd);
+    }
+
     msg = dbus_message_new_method_call(
         IFACE, "/org/freedesktop/Notifications",
         IFACE, "Notify");
@@ -626,106 +826,46 @@ main(int argc, char *const *argv)
     const char *icon_arg = icon != NULL ? icon : "";
     DBusMessageIter args;
     dbus_message_iter_init_append(msg, &args);
-    dbus_message_iter_append_basic(&args, 's', &app_id);
-    dbus_message_iter_append_basic(&args, 'u', &replaces_id);
-    dbus_message_iter_append_basic(&args, 's', &icon_arg);
-    dbus_message_iter_append_basic(&args, 's', &title);
-    dbus_message_iter_append_basic(&args, 's', &body);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &app_id);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &replaces_id);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &icon_arg);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &title);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &body);
 
     /* Actions: array of strings. Every even item is the action name,
        every odd item is the action label */
     DBusMessageIter args_actions;
-    dbus_message_iter_open_container(&args, 'a', "s", &args_actions);
+    dbus_message_iter_open_container(
+        &args, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING_AS_STRING, &args_actions);
     for (size_t i = 0; i < action_count; i++) {
         const struct action *a = &actions[i];
-        dbus_message_iter_append_basic(&args_actions, 's', &a->name);
-        dbus_message_iter_append_basic(&args_actions, 's', &a->label);
+        dbus_message_iter_append_basic(&args_actions, DBUS_TYPE_STRING, &a->name);
+        dbus_message_iter_append_basic(&args_actions, DBUS_TYPE_STRING, &a->label);
     }
     dbus_message_iter_close_container(&args, &args_actions);
 
-    /* Hints: dictionary of "sv" (string + variant) */
     DBusMessageIter args_hints;
-    dbus_message_iter_open_container(&args, 'a', "{sv}", &args_hints);
+    dbus_message_iter_open_container(
+        &args, DBUS_TYPE_ARRAY,
+        DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+          DBUS_TYPE_STRING_AS_STRING
+          DBUS_TYPE_VARIANT_AS_STRING
+        DBUS_DICT_ENTRY_END_CHAR_AS_STRING,
+        &args_hints);
     {
-        DBusMessageIter args_hints_hint;
-        DBusMessageIter args_hints_value;
-
-        dbus_message_iter_open_container(&args_hints, 'e', 0, &args_hints_hint);
-        {
-            static const char *hint_name = "urgency";
-            dbus_message_iter_append_basic(&args_hints_hint, 's', &hint_name);
-            dbus_message_iter_open_container(&args_hints_hint, 'v', "y", &args_hints_value);
-            {
-                dbus_message_iter_append_basic(&args_hints_value, 'y', &urgency_value);
-            }
-            dbus_message_iter_close_container(&args_hints_hint, &args_hints_value);
-        }
-        dbus_message_iter_close_container(&args_hints, &args_hints_hint);
-
-        if (icon != NULL) {
-            dbus_message_iter_open_container(&args_hints, 'e', 0, &args_hints_hint);
-            {
-                static const char *hint_name = "image-path";
-                dbus_message_iter_append_basic(&args_hints_hint, 's', &hint_name);
-                dbus_message_iter_open_container(&args_hints_hint, 'v', "s", &args_hints_value);
-                {
-                    if (icon_uri != NULL)
-                        dbus_message_iter_append_basic(&args_hints_value, 's', &icon_uri);
-                    else
-                    dbus_message_iter_append_basic(&args_hints_value, 's', &icon);
-                }
-                dbus_message_iter_close_container(&args_hints_hint, &args_hints_value);
-            }
-            dbus_message_iter_close_container(&args_hints, &args_hints_hint);
-        }
-
-        if (category != NULL) {
-            dbus_message_iter_open_container(&args_hints, 'e', 0, &args_hints_hint);
-            {
-                static const char *hint_name = "category";
-                dbus_message_iter_append_basic(&args_hints_hint, 's', &hint_name);
-                dbus_message_iter_open_container(&args_hints_hint, 'v', "s", &args_hints_value);
-                {
-                    dbus_message_iter_append_basic(&args_hints_value, 's', &category);
-                }
-                dbus_message_iter_close_container(&args_hints_hint, &args_hints_value);
-            }
-            dbus_message_iter_close_container(&args_hints, &args_hints_hint);
-        }
-
-        if (transient) {
-            dbus_message_iter_open_container(&args_hints, 'e', 0, &args_hints_hint);
-            {
-                static const char *hint_name = "transient";
-                dbus_message_iter_append_basic(&args_hints_hint, 's', &hint_name);
-                dbus_message_iter_open_container(&args_hints_hint, 'v', "b", &args_hints_value);
-                {
-                    dbus_message_iter_append_basic(&args_hints_value, 'b', &transient_value);
-                }
-                dbus_message_iter_close_container(&args_hints_hint, &args_hints_value);
-            }
-            dbus_message_iter_close_container(&args_hints, &args_hints_hint);
-        }
+        add_urgency_hint(&args_hints, urgency);
+        add_category_hint(&args_hints, category);
+        add_transient_hint(&args_hints, transient);
+        add_image_data_hint(&args_hints, image_data, image_width, image_height);
 
         /* User specified hints */
-        for (size_t i = 0; i < hint_count; i++) {
-            const struct hint *h = &hints[i];
-            dbus_message_iter_open_container(&args_hints, 'e', 0, &args_hints_hint);
-            {
-                dbus_message_iter_append_basic(&args_hints_hint, 's', &h->name);
-                dbus_message_iter_open_container(&args_hints_hint, 'v', h->type_as_string, &args_hints_value);
-                {
-                    dbus_message_iter_append_basic(&args_hints_value, h->type, &h->u);
-                }
-                dbus_message_iter_close_container(&args_hints_hint, &args_hints_value);
-            }
-            dbus_message_iter_close_container(&args_hints, &args_hints_hint);
-        }
+        for (size_t i = 0; i < hint_count; i++)
+            add_user_hint(&args_hints, &hints[i]);
     }
     dbus_message_iter_close_container(&args, &args_hints);
 
     /* Expire timeout */
-    dbus_message_iter_append_basic(&args, 'i', &expire_time);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_INT32, &expire_time);
 
     /* Sign up for signals *before* we send the notification, to avoid race */
     if (wait) {
@@ -897,6 +1037,7 @@ out:
         dbus_connection_unref(conn);
 
     dbus_error_free(&err);
+    free(image_data);
     free(icon_uri);
     free(body);
     free(hints);
